import bilby
from typing import TYPE_CHECKING
import numpy as np
import pandas as pd
import torch
from ..architectures.flows import FlowArchitecture
from ..callbacks import (
    StrainVisualization,
    SavePosterior,
    CrossMatchStatistics,
    ProbProbPlot,
    PlotMollview,
    PlotCorner,
    SaveFITS,
    SaveInjectionParameters,
)
from ...utils.result import AmplfiResult
from .base import AmplfiModel
from typing import Optional
from bilby.core.prior import PriorDict

if TYPE_CHECKING:
    from pathlib import Path
    from ..prior import AmplfiPrior

Tensor = torch.Tensor


class FlowModel(AmplfiModel):
    """
    A LightningModule for training normalizing flows

    Args:
        *args:
            See arguments in `amplfi.train.models.base.AmplfiModel`
        arch:
            Neural network architecture to train.
            This should be a subclass of `FlowArchitecture`.
        filter_params:
            If `True`, filter the samples produced by the flow
            to keep only valid samples within the prior boundaries.
        samples_per_event:
            Number of samples to draw per event for testing
        nside:
            Healpix nside used when creating skymaps
        min_samples_per_pix:
        num_plot:
            Number of testing events to plot skymaps, corner
            plots and, if `plot_data` is `True`, strain data
            visualizations.
        plot_data:
            If `True`, plot strain visualization for `num_plot`
            of the testing set events
        plot_corner:
            If `True`, plot corner plots for
            testing set events
        plot_mollview:
            If `True`, plot mollview plots for
            testing set events
        cross_match:
            If `True`, run ligo.skymap.postprocess.crossmatch
            on result objects at the end of testing epoch
            and produce searched area and volume cdfs
        save_fits:
            If `True`, save skymaps as FITS files
            for testing set events
        save_posterior:
            If `True`, save bilby Result objects and posterior samples
        save_injection_parameters:
            If `True`, save the injection parameters for each event
            in the testing set to a an hdf5 file. Useful for
            testing datasets where the injection parameters are randomly
            sampled.
        target_prior:
            Path to a bilby prior file for reweighting posterior samples to
            a new prior.
    """

    def __init__(
        self,
        *args,
        arch: FlowArchitecture,
        filter_params: bool = True,
        samples_per_event: int = 10000,
        nside: int = 32,
        min_samples_per_pix: int = 5,
        num_plot: int = 10,
        plot_data: bool = False,
        plot_corner: bool = True,
        plot_mollview: bool = True,
        cross_match: bool = True,
        save_fits: bool = True,
        save_posterior: bool = False,
        save_injection_parameters: bool = True,
        target_prior: Optional["Path"] = None,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.model = arch
        self.samples_per_event = samples_per_event
        self.num_plot = num_plot
        self.nside = nside
        self.min_samples_per_pix = min_samples_per_pix
        self.plot_data = plot_data
        self.plot_corner = plot_corner
        self.plot_mollview = plot_mollview
        self.save_fits = save_fits
        self.save_posterior = save_posterior
        self.cross_match = cross_match
        self.save_injection_parameters = save_injection_parameters
        self.filter_params = filter_params

        if target_prior is not None:
            target_prior = PriorDict(filename=str(target_prior))
        self.target_prior: Optional[PriorDict] = target_prior

        # save our hyperparameters
        self.save_hyperparameters(ignore=["arch"])

    def forward(self, context, parameters) -> Tensor:
        return -self.model.log_prob(parameters, context=context)

    def training_step(self, batch, _):
        strain, asds, parameters, _ = batch
        context = (strain, asds)
        loss = self(context, parameters).mean()
        self.log(
            "train_loss",
            loss,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            sync_dist=True,
            logger=True,
        )
        return loss

    def validation_step(self, batch, _):
        strain, asds, parameters, _ = batch
        context = (strain, asds)
        loss = self(context, parameters).mean()
        self.log(
            "valid_loss",
            loss,
            on_step=False,
            on_epoch=True,
            prog_bar=True,
            sync_dist=True,
            logger=True,
        )
        return loss

    def on_predict_epoch_start(self):
        self.on_test_epoch_start()

    def on_test_epoch_start(self):
        self.test_outdir.mkdir(exist_ok=True, parents=True)
        self.test_results: list[AmplfiResult] = []
        self.reweighted_results: list[AmplfiResult] = []

        # update the training prior to now include
        # the extrinisc parameters, so log probabilites can be calculated
        waveform_sampler = self.trainer.datamodule.waveform_sampler
        training_prior = waveform_sampler.training_prior

        for key in ["dec", "psi", "phi"]:
            training_prior.priors[key] = getattr(self.trainer.datamodule, key)

        self.training_prior: "AmplfiPrior" = training_prior

        # if reweighting, write target prior to test directory
        if self.target_prior is not None:
            self.target_prior.to_file(self.test_outdir, label="reweight")
            (self.test_outdir / "reweighted").mkdir(exist_ok=True)

    def on_test_batch_end(self, outputs, *_):
        result: "AmplfiResult"
        reweighted: Optional["AmplfiResult"]
        result, reweighted = outputs
        self.test_results.append(result)

        if reweighted is not None:
            self.reweighted_results.append(reweighted)

    def analyze_event(self, strain, asds, parameters=None, snr=None):
        context = (strain, asds)
        samples = self.model.sample(
            self.hparams.samples_per_event, context=context
        )
        log_probs = self.model.log_prob(samples, context)

        samples = samples.squeeze(1)
        log_probs = log_probs.squeeze(1)

        descaled = self.scale(samples, reverse=True)
        if self.filter_params:
            descaled, mask = self.filter_parameters(descaled)
            log_probs = log_probs[mask]

        # convert samples to dictionary for
        # calculating log probabilites
        samples_dict = dict(
            zip(
                self.hparams.inference_params,
                descaled.transpose(1, 0),
                strict=True,
            )
        )

        # calculate training prior probability of posterior samples
        log_prior_of_posterior_samples = self.training_prior.log_prob(
            samples_dict
        )

        # when predicting, there will be no ground truth parameters
        injection_parameters = None
        if parameters is not None:
            parameters = self.scale(parameters, reverse=True)
            parameters = parameters.cpu().numpy()[0]

            # create a dictionary of injection parameters
            # mapping from parameter string to the true
            # injection value, and add snr if provided
            injection_parameters = {
                k: float(v)
                for k, v in zip(
                    self.inference_params, parameters, strict=False
                )
            }
            injection_parameters["ra"] = injection_parameters["phi"]

            if snr is not None:
                injection_parameters["snr"] = snr[0].item()
        result = self.cast_as_bilby_result(
            descaled.cpu().numpy(),
            log_probs.cpu().numpy(),
            log_prior_of_posterior_samples.cpu().numpy(),
            injection_parameters,
        )

        # optionally reweight to different prior
        # TODO: include likelihood reweighting
        reweighted_result: Optional[AmplfiResult] = None
        if self.target_prior is not None:
            self._logger.info(
                f"Reweighting {len(result.posterior)} posterior samples"
            )
            reweighted_result = result.reweight_to_prior(self.target_prior)
            self._logger.info(
                f"{len(reweighted_result.posterior)} posterior samples "
                "after rejection sampling"
            )

        return result, reweighted_result

    def test_step(self, batch, _) -> AmplfiResult:
        strain, asds, parameters, snr = batch
        return self.analyze_event(strain, asds, parameters, snr)

    def predict_step(self, batch, _):
        strain, asds, _ = batch
        return self.analyze_event(strain, asds, None, None)

    def cast_as_bilby_result(
        self,
        samples: np.ndarray,
        log_probs: np.ndarray,
        log_prior_probs: np.ndarray,
        injection_parameters: Optional[dict[str, float]] = None,
    ) -> AmplfiResult:
        """Cast posterior samples as Bilby Result object
        for ease of producing corner and pp plots

        Args:
            samples:
                An array of posterior samples of shape
                (1, num_samples, num_params)
            log_probs:
                An array of log probabilities of posterior samples
                as predicted under the normalizing flow model
            log_prior_probs:
                An array of log prior probabilities of posterior samples
                as predicted under the training prior
            injection_parameters:
                For injections, a dictionary mapping from parameter string
                to the true injection value

        """

        # create dummy prior with correct attributes
        # for making our results compatible with bilbys make_pp_plot
        priors = {
            param: bilby.core.prior.base.Prior(latex_label=param)
            for param in self.inference_params
        }
        posterior = {}
        for idx, k in enumerate(self.inference_params):
            posterior[k] = samples.T[idx].flatten()

        posterior["log_prob"] = log_probs
        posterior = pd.DataFrame(posterior)

        r = AmplfiResult(
            label="PEModel",
            injection_parameters=injection_parameters,
            posterior=posterior,
            search_parameter_keys=self.inference_params,
            priors=priors,
            log_prior_evaluations=log_prior_probs,
        )
        r.posterior["ra"] = r.posterior["phi"]
        r.posterior["log_prior"] = log_prior_probs
        return r

    def filter_parameters(self, parameters: torch.Tensor):
        """
        Filter the descaled parameters to keep only valid samples
        within their boundaries.

        Args:
            descaled (torch.Tensor): The descaled parameters tensor.

        Returns:
            torch.Tensor: The filtered descaled parameters.
        """
        net_mask = torch.ones(
            parameters.shape[0], dtype=bool, device=parameters.device
        )
        priors = self.training_prior.priors
        for i, param in enumerate(self.inference_params):
            samples = parameters[:, i]
            if param in ["dec", "psi", "phi"]:
                prior = getattr(self.trainer.datamodule, param)
            else:
                prior = priors[param]

            mask = (prior.log_prob(samples) == float("-inf")).to(
                samples.device
            )
            self._logger.debug(
                f"Removed {mask.sum()}/{len(mask)} samples for parameter "
                f"{param} outside of prior range"
            )

            net_mask &= ~mask

        self._logger.info(
            f"Removed {(~net_mask).sum()}/{len(net_mask)} total samples "
            "outside of prior range"
        )
        parameters = parameters[net_mask]

        return parameters, net_mask

    def configure_callbacks(self):
        callbacks = super().configure_callbacks()
        callbacks += [ProbProbPlot()]

        event_outdir = self.test_outdir / "events"
        event_outdir.mkdir(parents=True, exist_ok=True)

        if self.plot_data:
            callbacks.append(StrainVisualization(event_outdir, self.num_plot))

        if self.save_injection_parameters:
            callbacks.append(SaveInjectionParameters(self.test_outdir))

        if self.save_fits:
            callbacks.append(
                SaveFITS(event_outdir, self.nside, self.min_samples_per_pix)
            )

        if self.plot_mollview:
            callbacks.append(PlotMollview(event_outdir, self.nside))

        if self.plot_corner:
            callbacks.append(PlotCorner(event_outdir))

        if self.save_posterior:
            callbacks.append(SavePosterior(event_outdir))

        if self.cross_match:
            callbacks.append(CrossMatchStatistics())

        return callbacks



from tempfile import NamedTemporaryFile
import pytest

import numpy as np

from amplfi.utils import skymap


@pytest.fixture(params=[32, 64])
def n_side(request):
    return request.param


@pytest.fixture(params=[10000, 20000])
def n_samples(request):
    return request.param


def test_histogram_skymap(n_side, n_samples):
    # mock posterior samples
    ra_samples = np.random.uniform(0, 2 * np.pi, n_samples)
    dec_samples = np.random.uniform(-0.5 * np.pi, 0.5 * np.pi, n_samples)
    dist_samples = np.random.uniform(1000, 2000, n_samples)
    # check skymap format with no extra metadata
    skymap_content = skymap.histogram_skymap(
        ra_samples, dec_samples, dist_samples, nside=n_side
    )
    # check FITS format
    with NamedTemporaryFile(mode="w+", suffix=".multiorder.fits") as fits_file:
        skymap_content.write(fits_file.name, format="fits", overwrite=True)
        from astropy.table import QTable

        t = QTable.read(fits_file.name)
    # do basic checks on the format
    assert t.meta["PIXTYPE"] == "HEALPIX"
    assert t.meta["ORDERING"] == "NUNIQ"

    # check if extra metadata is correctly added
    skymap_content = skymap.histogram_skymap(
        ra_samples,
        dec_samples,
        dist_samples,
        nside=n_side,
        metadata={"INSTRUME": "H1,L1,V1"},
    )
    with NamedTemporaryFile(mode="w+", suffix=".multiorder.fits") as fits_file:
        skymap_content.write(fits_file.name, format="fits", overwrite=True)
        from astropy.table import QTable

        t = QTable.read(fits_file.name)
    # do basic checks on the format
    assert t.meta["PIXTYPE"] == "HEALPIX"
    assert t.meta["ORDERING"] == "NUNIQ"
    assert t.meta["INSTRUME"] == "H1,L1,V1"



import bilby
from . import skymap
from typing import Optional
from pathlib import Path
import healpy as hp
from astropy import units as u
from astropy.table import Table
from astropy.coordinates import SkyCoord
import matplotlib.pyplot as plt
import numpy as np
from ligo.skymap.postprocess.crossmatch import crossmatch, CrossmatchResult
from copy import copy
import pandas as pd


class AmplfiResult(bilby.result.Result):
    """
    A subclass of `bilby.result.Result` with additional
    convenience methods for generating AMPLFI skymaps
    """

    def to_crossmatch_result(
        self,
        nside: int,
        min_samples_per_pix: int = 5,
        use_distance: bool = True,
        **crossmatch_kwargs,
    ) -> CrossmatchResult:
        """
        Calculate a `ligo.skymap.postprocess.crossmatch.CrossmatchResult`
        based on sky localization and distance posterior samples.
        The posterior dataframe and injection_parameters dict
        should have `ra` and `dec` entries

        """
        skymap = self.to_skymap(
            nside=nside,
            min_samples_per_pix=min_samples_per_pix,
            use_distance=use_distance,
        )

        coordinates = SkyCoord(
            self.injection_parameters["ra"] * u.rad,
            self.injection_parameters["dec"] * u.rad,
            distance=self.injection_parameters["distance"] * u.Mpc,
        )
        return crossmatch(skymap, coordinates, **crossmatch_kwargs)

    def to_skymap(
        self,
        nside: int,
        min_samples_per_pix: int = 5,
        use_distance: bool = True,
    ) -> Table:
        """
        Calculate a histogram skymap from posterior samples
        The posterior dataframe and injection_parameters dict
           should have `ra` and `dec` entries
        """
        distance = None
        if use_distance:
            distance = self.posterior["distance"]

        return skymap.histogram_skymap(
            self.posterior["ra"],
            self.posterior["dec"],
            distance,
            nside=nside,
            min_samples_per_pix=min_samples_per_pix,
        )

    def calculate_searched_area(
        self, nside: int = 32
    ) -> tuple[float, float, float]:
        """
        Calculate the searched area, and estimates
        of 50% and 90% credible region.
        The posterior dataframe and injection_parameters dict
        should have `ra` and `dec` entries
        """
        smap = self.to_skymap(nside)["PROBDENSITY"]

        return skymap.calculate_searched_area(
            smap,
            self.injection_parameters["ra"],
            self.injection_parameters["dec"],
            nside=nside,
        )

    def plot_mollview(
        self, nside: int = 32, outpath: Optional[Path] = None
    ) -> plt.Figure:
        """
        Plot a mollweide projection of the skymap.
        The posterior dataframe and injection_parameters dict
        should have `ra` and `dec` entries
        """
        skymap = self.to_skymap(nside)["PROBDENSITY"]

        # plot molleweide
        plt.close()
        fig = hp.mollview(skymap, nest=True)

        # plot true values if available
        if self.injection_parameters is not None:
            ra_inj = self.injection_parameters["phi"]
            dec_inj = self.injection_parameters["dec"]
            theta_inj = np.pi / 2 - dec_inj

            hp.visufunc.projscatter(
                theta_inj, ra_inj, marker="x", color="red", s=150
            )

        plt.savefig(outpath)

        return fig

    def reweight_to_prior(
        self,
        target_prior: bilby.core.prior.PriorDict,
        max_rejection_steps: int = 1,
    ) -> "AmplfiResult":
        reweighted_result = copy(self)
        keys = list(target_prior.keys())
        samples_dict = {key: self.posterior[key].values for key in keys}
        target_probs = target_prior.ln_prob(samples_dict, axis=0)
        ln_weights = target_probs - self.posterior["log_prior"].values
        weights = np.exp(ln_weights)

        num_samples = len(self.posterior)
        reweighted_posterior = pd.DataFrame(columns=self.posterior.columns)
        for _ in range(max_rejection_steps):
            if len(reweighted_posterior) >= num_samples:
                break

            samples = bilby.core.result.rejection_sample(
                self.posterior, weights
            )
            reweighted_posterior = pd.concat(
                [reweighted_posterior, samples], ignore_index=True
            )

        reweighted_result.posterior = reweighted_posterior[:num_samples]
        return reweighted_result



from functools import partial
import law
import numpy as np
import luigi
from ligo.skymap.postprocess.crossmatch import crossmatch
from ligo.skymap.tool import ligo_skymap_plot
from astropy.coordinates import SkyCoord
from astropy import units as u
import h5py
from mldatafind.law.parameters import PathParameter
from mldatafind.law.tasks import Fetch
from mldatafind.law.tasks import Query as _Query
from mldatafind.law.tasks.condor.workflows import StaticMemoryWorkflow
from .base import DATA_SANDBOX, AmplfiDataTaskMixin
from .paths import paths
from luigi.util import inherits
from ligo.skymap.io.fits import read_sky_map
from tqdm.auto import tqdm
import multiprocessing as mp


# add mixin for appending amplfi specific
# environment variables to condor and containers
class Query(AmplfiDataTaskMixin, _Query):
    pass


# override FetchTrain and FetchTest
# tasks to use Query with mixin
class FetchTrain(AmplfiDataTaskMixin, Fetch):
    def workflow_requires(self):
        reqs = {}
        reqs["segments"] = Query.req(self, segments_file=self.segments_file)
        return reqs


class FetchTest(AmplfiDataTaskMixin, Fetch):
    def workflow_requires(self):
        reqs = {}
        reqs["segments"] = Query.req(self, segments_file=self.segments_file)
        return reqs


class DataGeneration(law.WrapperTask):
    """
    Pipeline for launching FetchTrain and FetchTest tasks
    """

    dev = luigi.BoolParameter(
        default=False, description="Run the task in development mode."
    )

    def requires(self):
        yield FetchTrain.req(
            self,
            sandbox=DATA_SANDBOX,
            data_dir=paths().data_dir / "train" / "background",
            condor_directory=paths().condor_dir / "train",
        )

        yield FetchTest.req(
            self,
            sandbox=DATA_SANDBOX,
            data_dir=paths().data_dir / "test" / "background",
            condor_directory=paths().condor_dir / "test",
        )


class LigoSkymap(
    AmplfiDataTaskMixin,
    law.LocalWorkflow,
    StaticMemoryWorkflow,
    law.SandboxTask,
):
    """
    Workflow for parallelizing skymap generation via ligo-skymap-from-samples
    """

    data_dir = PathParameter(
        description="Path to the directory containing the "
        "event sub directories."
        "Each sub directory should contain "
        "a posterior_samples.dat file."
    )
    ligo_skymap_args = luigi.OptionalListParameter(
        description="Additional command line style arguments"
        "to pass to ligo-skymap-from-samples.",
        default="",
    )
    dev = luigi.BoolParameter(
        default=False, description="Run the task in development mode."
    )
    sandbox = DATA_SANDBOX

    def sandbox_env(self, env):
        env = super().sandbox_env(env)
        env.update({"MKL_NUM_THREADS": "1", "OMP_NUM_THREADS": "1"})
        return env

    def create_branch_map(self):
        branch_map = {}
        num = len([x for x in self.data_dir.iterdir() if x.is_dir()])
        for idx in range(num):
            branch_map[idx] = (
                self.data_dir / str(idx) / "posterior_samples.dat"
            )
        return branch_map

    def output(self):
        event_dir = self.branch_data.parent
        return law.LocalFileTarget(event_dir / "skymap.fits")

    def run(self):
        from ligo.skymap.tool import (
            ligo_skymap_from_samples,
        )

        args = [
            str(self.branch_data),
            "-j",
            str(self.request_cpus),
            "-o",
            str(self.branch_data.parent),
        ]
        if self.ligo_skymap_args:
            args.extend(self.ligo_skymap_args)

        # call ligo-skymap-from-samples
        ligo_skymap_from_samples.main(args)


CROSSMATCH_ATTRS = [
    "searched_area",
    "searched_vol",
    "searched_prob",
    "searched_prob_vol",
    "searched_prob_dist",
    "offset",
]


def process_skymap(skymap_item, parameter_file, plot, data_dir):
    i, skymap_target = skymap_item
    skymap = read_sky_map(skymap_target.path, moc=True, distances=True)

    with h5py.File(parameter_file, "r") as f:
        ra = f["phi"][i] * u.rad
        dec = f["dec"][i] * u.rad
        dist = f["distance"][i] * u.Mpc

    coord = SkyCoord(ra=ra, dec=dec, distance=dist)

    if plot:
        ligo_skymap_plot.main(
            [
                skymap_target.path,
                "--radec",
                str(ra.to(u.deg).value),
                str(dec.to(u.deg).value),
                "--output",
                str(data_dir / str(i) / "ligo_skymap.png"),
            ]
        )

    # Perform crossmatch
    cm = crossmatch(skymap, coord)

    # Extract attributes
    result = {}
    for attr in CROSSMATCH_ATTRS:
        result[attr] = getattr(cm, attr)

    return i, result


@inherits(LigoSkymap)
class AggregateLigoSkymap(
    AmplfiDataTaskMixin,
    law.SandboxTask,
):
    """
    Task to calculate and aggregate statstics from
    the skymaps generated by the LigoSkymap using
    ligo.skymap.postprocess.crossmatch.
    """

    parameter_file = luigi.OptionalParameter(
        default="",
        description="Path to an hdf5 file containing `phi`, `dec` and `dist`"
        " datasets corresponding to the ground truth values of the event. For "
        "example, this can be the output of the "
        "`SaveInjectionParameters` lightning callback",
    )
    plot = luigi.BoolParameter(
        default=False,
        description="If True, will generate a plot using "
        "ligo.skymap.tool.ligo_skymap_plot",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sandbox = DATA_SANDBOX

    def requires(self):
        return LigoSkymap.req(
            self,
            request_disk=self.request_disk,
            request_cpus=self.request_cpus,
            request_memory=self.request_memory,
        )

    def output(self):
        return law.LocalFileTarget(self.data_dir / "ligo_skymap_stats.hdf5")

    def run(self):
        func = partial(
            process_skymap,
            parameter_file=self.parameter_file,
            plot=self.plot,
            data_dir=self.data_dir,
        )
        skymaps = self.input()["collection"].targets
        with mp.Pool(processes=min(mp.cpu_count(), len(skymaps))) as pool:
            results = list(
                tqdm(
                    pool.imap(func, skymaps.items()),
                    total=len(skymaps),
                    desc="Crossmatching skymaps",
                )
            )

        data = {attr: np.zeros(len(skymaps)) for attr in CROSSMATCH_ATTRS}

        for i, result in results:
            for attr in CROSSMATCH_ATTRS:
                data[attr][i] = result[attr]

        with h5py.File(self.output().path, "w") as f:
            for attr, values in data.items():
                f.create_dataset(attr, data=values)



import logging
import torch
from functools import partial
from concurrent.futures import ProcessPoolExecutor, as_completed
import multiprocessing as mp
from astropy import io
from pathlib import Path
from typing import TYPE_CHECKING, Optional
from gwpy.plot import Plot
from gwpy.timeseries import TimeSeries
import lightning.pytorch as pl
import matplotlib.pyplot as plt
import numpy as np
import scipy
import h5py
import bilby
from tqdm.auto import tqdm
import ligo.skymap.plot  # noqa: F401

if TYPE_CHECKING:
    from ligo.skymap.postprocess.crossmatch import CrossmatchResult
    from amplfi.train.models import FlowModel
    from amplfi.utils.result import AmplfiResult


class StrainVisualization(pl.Callback):
    """
    Lightning Callback for visualizing the strain data
    and asds being analyzed during the test step
    """

    def __init__(self, outdir: Path, num_plot: int, save_data: bool = True):
        self.outdir = outdir
        self.num_plot = num_plot
        self.save_data = save_data

    def plot_strain(self, outdir, result, batch, trainer):
        """
        Called at the end of each test step.
        `outputs` consists of objects returned by `pl_module.test_step`.
        """

        # unpack batch
        strain, asds, *_ = batch
        strain, asds = strain[0].cpu().numpy(), asds[0].cpu().numpy()

        # steal some attributes needed from datamodule
        # TODO: should we link these to the pl_module?
        highpass = trainer.datamodule.hparams.highpass
        sample_rate = trainer.datamodule.hparams.sample_rate
        ifos = trainer.datamodule.hparams.ifos

        # filenames for various plots
        whitened_td_strain_fname = outdir / "whitened_td.png"
        whitened_fd_strain_fname = outdir / "whitened_fd.png"
        qscan_fname = outdir / "qscan.png"
        asd_fname = outdir / "asd.png"

        # whitened time domain strain
        plt.figure()
        plt.title("Whitened Time Domain Strain")

        for i, ifo in enumerate(ifos):
            plt.plot(strain[i], label=ifo)
            if self.save_data:
                np.savetxt(outdir / "whitened_td_strain.txt", strain)

        plt.legend()
        plt.savefig(whitened_td_strain_fname)
        plt.close()

        # qscans
        qscans = []

        for i in range(len(ifos)):
            ts = TimeSeries(
                strain[i],
                dt=1 / sample_rate,
            )

            spec = ts.q_transform(
                logf=True,
                whiten=False,
                frange=(25, 500),
                qrange=(4, 108),
                # outseg=(3, 3.7),
            )
            qscans.append(spec)

        title = ""
        if result.injection_parameters is not None:
            chirp_mass, mass_ratio = (
                result.injection_parameters["chirp_mass"],
                result.injection_parameters["mass_ratio"],
            )
            title = f"chirp_mass: {chirp_mass:2f}, mass_ratio: {mass_ratio:2f}"
        plot = Plot(
            *qscans,
            figsize=(18, 5),
            geometry=(1, len(ifos)),
            yscale="log",
            method="pcolormesh",
            cmap="viridis",
            title=title,
        )

        for i, ax in enumerate(plot.axes):
            label = "" if i != 1 else "Normalized Energy"
            plot.colorbar(ax=ax, label=label)

        plot.savefig(qscan_fname)
        plt.close()

        # asds
        frequencies = trainer.datamodule.frequencies.numpy()
        mask = frequencies > highpass
        frequencies_masked = frequencies[mask]
        plt.figure()
        for i, ifo in enumerate(ifos):
            plt.loglog(frequencies_masked, asds[i], label=f"{ifo} asd")
            plt.title("ASDs")
            plt.xlabel("Frequency (Hz)")
            plt.ylabel("Scaled Amplitude")

        if self.save_data:
            np.savetxt(
                outdir / "asds.txt",
                np.vstack((np.expand_dims(frequencies_masked, axis=0), asds)),
            )
        plt.legend()
        plt.savefig(asd_fname)
        plt.close()

        # whitened frequency domain data
        fig, axes = plt.subplots(1, 2, figsize=(14, 6))
        for i, ifo in enumerate(ifos):
            strain_fft = np.fft.rfft(strain[i])
            freqs = np.fft.rfftfreq(n=strain[i].shape[-1], d=1 / sample_rate)

            axes[0].plot(
                freqs,
                strain_fft.real / (sample_rate ** (1 / 2)),
                label=f"{ifo} data",
            )
            axes[1].plot(
                freqs,
                strain_fft.imag / (sample_rate ** (1 / 2)),
                label=f"{ifo} data",
            )

        axes[0].set_title("real")
        axes[1].set_title("imaginary")

        axes[0].set_ylabel("whitened amplitude")
        axes[0].set_xlabel("frequency (hz)")
        axes[0].set_xlabel("frequency (hz)")

        plt.legend()

        plt.savefig(whitened_fd_strain_fname)
        plt.close()

    def on_test_batch_end(
        self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx=0
    ) -> None:
        # test_step returns bilby result object
        # and optionally a reweighted result
        result: "AmplfiResult"
        _: Optional["AmplfiResult"]
        result, _ = outputs

        if batch_idx >= self.num_plot:
            return

        outdir = self.outdir / str(batch_idx)
        outdir.mkdir(exist_ok=True)
        self.plot_strain(outdir, result, batch, trainer)

    def on_predict_batch_end(
        self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx=0
    ):
        # test_step returns bilby result object
        # and optionally a reweighted result
        result: "AmplfiResult"
        _: Optional["AmplfiResult"]
        result, _ = outputs

        if batch_idx >= self.num_plot:
            return

        gpstime = batch[2].cpu().numpy()[0]
        outdir = self.outdir / str(int(gpstime))
        outdir.mkdir(exist_ok=True)
        self.plot_strain(outdir, result, batch, trainer)


class PlotMollview(pl.Callback):
    """
    Lightning Callback for plotting mollview skymaps
    after each test batch
    """

    def __init__(self, outdir: Path, nside: int):
        self.outdir = outdir
        self.nside = nside

    def on_test_batch_end(
        self,
        trainer,
        pl_module: "FlowModel",
        outputs: "AmplfiResult",
        batch,
        batch_idx,
        dataloader_idx=0,
    ) -> None:
        """
        Called at the end of each test step.
        `outputs` consists of objects returned by `pl_module.test_step`.
        """

        # test_step returns bilby result object
        # and optionally a reweighted result
        result: "AmplfiResult"
        reweighted: Optional["AmplfiResult"]
        result, reweighted = outputs

        outdir = self.outdir / str(batch_idx)
        outdir.mkdir(exist_ok=True)
        skymap_filename = outdir / "mollview.png"
        result.plot_mollview(
            self.nside,
            outpath=skymap_filename,
        )

        if reweighted is not None:
            reweighted_outdir = outdir / "reweighted"
            skymap_filename = reweighted_outdir / "mollview.png"
            reweighted_outdir.mkdir(exist_ok=True)
            reweighted.plot_mollview(
                self.nside,
                outpath=skymap_filename,
            )

    def on_predict_batch_end(
        self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx=0
    ):
        # test_step returns bilby result object
        # and optionally a reweighted result
        result: "AmplfiResult"
        reweighted: Optional["AmplfiResult"]
        result, reweighted = outputs

        # TODO: remove "magic" index 2
        gpstime = batch[2].cpu().numpy()[0]
        outdir = self.outdir / str(int(gpstime))
        outdir.mkdir(exist_ok=True)
        skymap_filename = outdir / "mollview.png"
        result.plot_mollview(
            self.nside,
            outpath=skymap_filename,
        )

        if reweighted is not None:
            reweighted_outdir = outdir / "reweighted"
            skymap_filename = reweighted_outdir / "mollview.png"
            reweighted_outdir.mkdir(exist_ok=True)
            reweighted.plot_mollview(
                self.nside,
                outpath=skymap_filename,
            )


class PlotCorner(pl.Callback):
    """
    Lightning Callback for making corner plots after test epoch
    """

    def __init__(self, outdir: Path):
        self.outdir = outdir

    def plot_corner(self, result: "AmplfiResult", outdir: Path):
        corner_filename = outdir / "corner.png"
        outdir.mkdir(exist_ok=True)
        result.plot_corner(
            save=True,
            filename=corner_filename,
            levels=(0.5, 0.9),
        )

    def on_test_batch_end(
        self,
        trainer,
        pl_module: "FlowModel",
        outputs: "AmplfiResult",
        batch,
        batch_idx,
        dataloader_idx=0,
    ) -> None:
        """
        Called at the end of each test step.
        `outputs` consists of objects returned by `pl_module.test_step`.
        """

        # test_step returns bilby result object
        # and optionally a reweighted result
        result: "AmplfiResult"
        reweighted: Optional["AmplfiResult"]
        result, reweighted = outputs

        outdir = self.outdir / str(batch_idx)
        self.plot_corner(result, outdir)

        if reweighted is not None:
            outdir = outdir / "reweighted"
            self.plot_corner(reweighted, outdir)

    def on_predict_batch_end(
        self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx=0
    ):
        # test_step returns bilby result object
        # and optionally a reweighted result
        result: "AmplfiResult"
        reweighted: Optional["AmplfiResult"]
        result, reweighted = outputs

        # for predict step, use gpstime to name the directory
        gpstime = batch[2].cpu().numpy()[0]
        outdir = self.outdir / str(int(gpstime))
        self.plot_corner(result, outdir)

        if reweighted is not None:
            # for predict step, use gpstime to name the directory
            outdir = outdir / "reweighted"
            self.plot_corner(reweighted, outdir)


class SaveFITS(pl.Callback):
    """ """

    def __init__(self, outdir: Path, nside: int, min_samples_per_pix: int):
        self.outdir = outdir
        self.nside = nside
        self.min_samples_per_pix = min_samples_per_pix

    def save_fits(
        self,
        result: "AmplfiResult",
        outdir: Path,
    ):
        fits = io.fits.table_to_hdu(
            result.to_skymap(
                self.nside,
                self.min_samples_per_pix,
                use_distance=True,
            )
        )
        outdir.mkdir(exist_ok=True)
        fits.writeto(outdir / "amplfi.skymap.fits", overwrite=True)

    def on_test_batch_end(
        self,
        trainer,
        pl_module,
        outputs: "AmplfiResult",
        batch,
        batch_idx,
        dataloader_idx=0,
    ) -> None:
        """
        Called at the end of each test step.
        `outputs` consists of objects returned by `pl_module.test_step`.
        """

        # test_step returns bilby result object
        # and optionally a reweighted result
        result: "AmplfiResult"
        reweighted: Optional["AmplfiResult"]
        result, reweighted = outputs

        outdir = self.outdir / str(batch_idx)
        outdir.mkdir(exist_ok=True)
        self.save_fits(result, outdir)

        if reweighted is not None:
            outdir = outdir / "reweighted"
            outdir.mkdir(exist_ok=True)
            self.save_fits(reweighted, outdir)

    def on_predict_batch_end(
        self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx=0
    ):
        result: "AmplfiResult"
        reweighted: Optional["AmplfiResult"]
        result, reweighted = outputs

        # for predict step, use gpstime to name the directory
        gpstime = batch[2].cpu().numpy()[0]
        outdir = self.outdir / str(int(gpstime))
        self.save_fits(result, outdir)

        if reweighted is not None:
            outdir = outdir / "reweighted"
            self.save_fits(reweighted, outdir)


class SavePosterior(pl.Callback):
    """
    Lightning Callback for bilby result objects and
    posterior data to disk
    """

    def __init__(self, outdir: Path):
        self.outdir = outdir

    def save_posterior(
        self,
        result: "AmplfiResult",
        outdir: Path,
    ):
        # save posterior samples for ease of use with
        # ligo skymap and save full result to have
        # access to the true injection parameters
        result.save_posterior_samples(outdir / "posterior_samples.dat")
        result.save_to_file(
            outdir / "result.hdf5", extension="hdf5", overwrite=True
        )

    def on_test_batch_end(
        self,
        trainer,
        pl_module,
        outputs: "AmplfiResult",
        batch,
        batch_idx,
        dataloader_idx=0,
    ) -> None:
        """
        Called at the end of each test step.
        `outputs` consists of objects returned by `pl_module.test_step`.
        """

        # test_step returns bilby result object
        # and optionally a reweighted result
        result: "AmplfiResult"
        reweighted: Optional["AmplfiResult"]
        result, reweighted = outputs

        outdir = self.outdir / str(batch_idx)
        outdir.mkdir(exist_ok=True)
        self.save_posterior(result, outdir)

        if reweighted is not None:
            outdir = outdir / "reweighted"
            outdir.mkdir(exist_ok=True)
            self.save_posterior(reweighted, outdir)

    def on_predict_batch_end(
        self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx=0
    ):
        # test_step returns bilby result object
        # and optionally a reweighted result
        result: "AmplfiResult"
        reweighted: Optional["AmplfiResult"]
        result, reweighted = outputs

        # for predict step, use gpstime to name the directory
        gpstime = batch[2].cpu().numpy()[0]

        outdir = self.outdir / str(int(gpstime))
        outdir.mkdir(exist_ok=True)
        self.save_posterior(result, outdir)

        if reweighted is not None:
            outdir = outdir / "reweighted"
            outdir.mkdir(exist_ok=True)
            self.save_posterior(reweighted, outdir)


class ProbProbPlot(pl.Callback):
    def on_test_epoch_end(self, _, pl_module: "FlowModel"):
        logger = pl_module._logger
        logger.info("Making PP plot")
        bilby.result.make_pp_plot(
            pl_module.test_results,
            save=True,
            filename=pl_module.test_outdir / "plots" / "pp-plot.png",
            keys=pl_module.inference_params,
        )

        if pl_module.reweighted_results:
            outdir = pl_module.test_outdir / "reweighted" / "plots"
            outdir.mkdir(exist_ok=True, parents=True)
            bilby.result.make_pp_plot(
                pl_module.reweighted_results,
                save=True,
                filename=outdir / "pp-plot.png",
                keys=pl_module.inference_params,
            )


def crossmatch_skymap(
    result: "AmplfiResult",
    nside: int,
    min_samples_per_pix: int,
    contours: tuple[float],
):
    """
    Function to process each skymap and calculate crossmatch statistics.
    """

    crossmatch_result = result.to_crossmatch_result(
        nside=nside,
        min_samples_per_pix=min_samples_per_pix,
        use_distance=True,
        contours=contours,
    )
    return crossmatch_result


class CrossMatchStatistics(pl.Callback):
    """
    Use `ligo.skymap.postprocess.crossmatch` to calculate skymap statistics.
    Make searched area and searched volume CDF plots
    """

    crossmatch_attributes = [
        "searched_area",
        "searched_vol",
        "searched_prob",
        "searched_prob_vol",
        "searched_prob_dist",
        "offset",
        "contour_areas",
    ]

    def __init__(self, contours: tuple[float] = (0.5, 0.9)):
        self.contours = contours

    def write_skymap_statistics(
        self, outdir: Path, results: list["CrossmatchResult"]
    ):
        """
        Write skymap statistics to file
        """

        outdir.mkdir(exist_ok=True)
        with h5py.File(outdir / "skymap_stats.hdf5", "w") as f:
            for attr in self.crossmatch_attributes:
                if attr == "contour_areas":
                    for i, contour in enumerate(self.contours):
                        f.create_dataset(
                            f"contour_areas_{int(contour * 100)}",
                            data=np.array(
                                [result.contour_areas[i] for result in results]
                            ),
                        )
                else:
                    f.create_dataset(
                        attr,
                        data=np.array(
                            [getattr(result, attr) for result in results]
                        ),
                    )

    def on_test_epoch_end(self, _, pl_module: "FlowModel"):
        self.crossmatch(
            pl_module.test_results,
            pl_module.test_outdir,
            pl_module.nside,
            pl_module.min_samples_per_pix,
        )

        if pl_module.reweighted_results:
            self.crossmatch(
                pl_module.reweighted_results,
                pl_module.test_outdir / "reweighted",
                pl_module.nside,
                pl_module.min_samples_per_pix,
            )

    def crossmatch(
        self,
        results: list["AmplfiResult"],
        outdir: Path,
        nside: int,
        min_samples_per_pix: int,
    ) -> None:
        (outdir / "plots").mkdir(exist_ok=True)
        func = partial(
            crossmatch_skymap,
            nside=nside,
            min_samples_per_pix=min_samples_per_pix,
            contours=self.contours,
        )

        crossmatch_results: list["CrossmatchResult"] = [None] * len(results)
        num_processes = min(mp.cpu_count(), len(results))
        with ProcessPoolExecutor(max_workers=num_processes) as executor:
            future_to_index = {
                executor.submit(func, result): idx
                for idx, result in enumerate(results)
            }
            for future in tqdm(
                as_completed(future_to_index),
                total=len(future_to_index),
                desc="Crossmatching skymaps",
            ):
                idx = future_to_index[future]
                crossmatch_results[idx] = future.result()

        self.write_skymap_statistics(outdir, crossmatch_results)

        # searched area cum hist
        searched_areas = [
            result.searched_area for result in crossmatch_results
        ]
        searched_volumes = [
            result.searched_vol for result in crossmatch_results
        ]
        fifty_percent_areas = [
            result.contour_areas[0] for result in crossmatch_results
        ]
        ninety_percent_areas = [
            result.contour_areas[1] for result in crossmatch_results
        ]

        searched_areas = np.sort(searched_areas)
        searched_volumes = np.sort(searched_volumes)
        counts = np.arange(1, len(searched_areas) + 1) / len(searched_areas)

        # searched volume cum hist
        plt.figure(figsize=(10, 6))
        plt.step(searched_volumes, counts, where="post")
        plt.xscale("log")
        plt.xlabel("Searched Volume (Mpc^3)")
        plt.ylabel("Cumulative Probability")
        plt.title("Searched Volume Cumulative Distribution Function")
        plt.grid()
        plt.axhline(0.5, color="grey", linestyle="--")
        plt.savefig(outdir / "plots" / "searched_volume.png")

        # searched area cum hist
        plt.figure(figsize=(10, 6))
        plt.step(searched_areas, counts, where="post")
        plt.xscale("log")
        plt.xlabel("Searched Area (deg^2)")
        plt.ylabel("Cumulative Probability")
        plt.title("Searched Area Cumulative Distribution Function")
        plt.grid()
        plt.axhline(0.5, color="grey", linestyle="--")
        plt.savefig(outdir / "plots" / "searched_area.png")

        plt.close()
        plt.figure(figsize=(10, 6))
        mm, bb, pp = plt.hist(
            fifty_percent_areas, label="50 percent area", bins=50
        )
        _, _, _ = plt.hist(
            ninety_percent_areas, label="90 percent area", bins=bb
        )
        plt.xlabel("Sq. deg.")
        plt.legend()
        plt.savefig(outdir / "plots" / "fifty_ninety_areas.png")
        plt.close()

        # searched prob pp-plot
        searched_probs = [
            result.searched_prob for result in crossmatch_results
        ]

        fig = plt.figure(figsize=(7, 7))
        ax = fig.add_subplot(111, projection="pp_plot")
        plt.rcParams.update({"font.size": 16})

        number_of_samples = len(searched_probs)
        alphas = [0.68, 0.95, 0.997]
        for alpha in alphas:
            ax.add_confidence_band(
                number_of_samples,
                alpha=alpha,
                color=(0, 0, 0, 0.1),
                edgecolor=(0, 0, 0, 0.2),
                annotate=False,
            )
        ax.add_diagonal()
        p = scipy.stats.kstest(searched_probs, "uniform").pvalue
        ax.add_series(
            searched_probs,
            label="AMPLFI"
            + r"$~({0:#.2g})$ ".format(round(p, 2))
            + str(len(searched_probs))
            + " events",
            color="saddlebrown",
            linewidth=2,
        )

        plt.title("Searched Area Probability-Probability Plot")
        ax.set_xlabel("Credible interval")
        ax.set_ylabel("Fraction of events in credible interval")
        ax.grid(True)
        ax.legend()
        fig.savefig(outdir / "plots" / "searched_prob_pp_plot.png")
        plt.close()

        # searched prob-vol pp-plot
        searched_prob_vols = [
            result.searched_prob_vol for result in crossmatch_results
        ]

        fig = plt.figure(figsize=(7, 7))
        ax = fig.add_subplot(111, projection="pp_plot")
        plt.rcParams.update({"font.size": 16})

        number_of_samples = len(searched_prob_vols)
        alphas = [0.68, 0.95, 0.997]
        for alpha in alphas:
            ax.add_confidence_band(
                number_of_samples,
                alpha=alpha,
                color=(0, 0, 0, 0.1),
                edgecolor=(0, 0, 0, 0.2),
                annotate=False,
            )
        ax.add_diagonal()
        p = scipy.stats.kstest(searched_prob_vols, "uniform").pvalue
        ax.add_series(
            searched_prob_vols,
            label="AMPLFI"
            + r"$~({0:#.2g})$ ".format(round(p, 2))
            + str(len(searched_prob_vols))
            + " events",
            color="saddlebrown",
            linewidth=2,
        )

        plt.title("Searched Volume Probability-Probability Plot")
        ax.set_xlabel("Credible interval")
        ax.set_ylabel("Fraction of events in credible interval")
        ax.grid(True)
        ax.legend()
        fig.savefig(outdir / "plots" / "searched_prob_vol_pp_plot.png")
        plt.close()


class SaveInjectionParameters(pl.Callback):
    """
    Save randomly sampled injection parameters to a file
    at the end of each test epoch.
    """

    def __init__(self, outdir: Path):
        self.outdir = outdir

    # TODO: should these parameters be saved
    # regardless of whether this callback is activated?
    # i.e. should `on_test_epoch_start` and `on_test_batch_end`
    # be moved to the base dataset?
    def on_test_epoch_start(self, trainer, pl_module: "FlowModel"):
        # initialize field in test parameters for snr
        num_test = len(trainer.datamodule.test_dataloader())
        for param in ["snr", "dec", "psi", "phi", "ra"]:
            trainer.datamodule.test_parameters[param] = np.zeros(num_test)

    def on_test_epoch_end(self, trainer, pl_module: "FlowModel"):
        # save parameters of randomly sampled injections
        with h5py.File(self.outdir / "parameters.hdf5", "w") as f:
            for param, data in trainer.datamodule.test_parameters.items():
                f.create_dataset(param, data=data)

    def on_test_batch_end(
        self,
        trainer,
        pl_module: "FlowModel",
        outputs,
        batch,
        batch_idx,
        dataloader_idx=0,
    ):
        result: "AmplfiResult"
        _: Optional["AmplfiResult"]
        result, _ = outputs
        for param in ["snr", "dec", "psi", "phi", "ra"]:
            trainer.datamodule.test_parameters[param][batch_idx] = (
                result.injection_parameters[param]
            )


class EstimateSamplingLatency(pl.Callback):
    """
    Estimate the sampling latency of the model by running
    a forward pass on a batch of data and measuring the time taken.
    """

    def __init__(self, num_samples: int = 20000, num_trials: int = 10):
        self.num_samples = num_samples
        self.num_trials = num_trials
        self.logger = logging.getLogger("EstimateSamplingLatency")

    def on_test_start(self, trainer, pl_module: "FlowModel"):
        batch = next(iter(trainer.datamodule.test_dataloader()))
        batch = trainer.datamodule.transfer_batch_to_device(
            batch, pl_module.device, 0
        )
        strain, asds, _, _ = trainer.datamodule.on_after_batch_transfer(
            batch, None
        )
        context = (strain, asds)
        times = []
        self.logger.info(
            "Estimating sampling latency by sampling "
            f" {self.num_samples} samples {self.num_trials} times..."
        )
        with torch.no_grad():
            for i in range(self.num_trials):
                start_time = torch.cuda.Event(enable_timing=True)
                end_time = torch.cuda.Event(enable_timing=True)

                start_time.record()
                _ = pl_module.model.sample(
                    self.num_samples,
                    context,
                )
                end_time.record()

                # wait for the events to be recorded
                torch.cuda.synchronize()

                # convert to seconds
                elapsed_time = start_time.elapsed_time(end_time) / 1000
                times.append(elapsed_time)
                self.logger.info(f"Trial {i + 1}: {elapsed_time:.2f} s")
        avg_time = np.mean(times)
        self.logger.info(f"Mean time: {avg_time:.2f} s")

    def on_train_start(self, trainer, pl_module):
        return self.on_test_start(trainer, pl_module)



import torch


class ZippedDataset(torch.utils.data.IterableDataset):
    def __init__(self, *datasets):
        super().__init__()
        self.datasets = datasets

    def __len__(self):
        lengths = []
        for dset in self.datasets:
            try:
                lengths.append(len(dset))
            except Exception as e:
                raise e from None
        return min(lengths)

    def __iter__(self):
        return zip(*self.datasets, strict=False)



from typing import Callable, Optional

import torch


class AmplfiPrior:
    def __init__(
        self,
        priors: dict[str, torch.distributions.Distribution],
        conversion_function: Optional[Callable] = None,
    ):
        """
        A class for sampling parameters from a prior distribution

        Args:
            priors:
                A dictionary of parameter samplers that take an integer N
                and return a tensor of shape (N, ...) representing
                samples from the prior distribution
            conversion_function:
                A callable that takes a dictionary of sampled parameters
                and returns a dictionary of waveform generation parameters
        """
        super().__init__()
        self.priors = priors
        self.conversion_function = conversion_function or (lambda x: x)

    def __call__(
        self,
        N: int,
        device: str = "cpu",
    ) -> dict[str, torch.Tensor]:
        """
        Generates random samples from the prior

        Args:
            N: Number of samples to generate
            device: Device to place the samples
        """
        # sample parameters from prior
        parameters = {
            k: v.sample((N,)).to(device) for k, v in self.priors.items()
        }
        # perform any necessary conversions
        # to from sampled parameters to
        # waveform generation parameters
        parameters = self.conversion_function(parameters)
        return parameters

    def log_prob(self, samples: dict[str, torch.Tensor]) -> torch.Tensor:
        """
        Calculate the log probability of samples under the prior

        Args:
            samples:
                Dictionary where key is parameter and
                value is tensor of samples
        """

        first = samples[list(samples.keys())[0]]
        log_probs = torch.ones(len(first), device=first.device)
        for param, tensor in samples.items():
            log_probs += self.priors[param].log_prob(tensor).to(first.device)
        return log_probs


class ParameterTransformer(torch.nn.Module):
    """
    Helper class for applying preprocessing
    transformations to inference parameters
    """

    def __init__(self, **transforms: Callable):
        super().__init__()
        self.transforms = transforms

    def forward(
        self,
        parameters: dict[str, torch.Tensor],
    ):
        # transform parameters
        transformed = {k: v(parameters[k]) for k, v in self.transforms.items()}
        # update parameter dict
        parameters.update(transformed)
        return parameters



import torch
from ml4gw.waveforms.conversion import (
    bilby_spins_to_lalsim,
    chirp_mass_and_mass_ratio_to_components,
)


def precessing_to_lalsimulation_parameters(
    parameters: dict[str, torch.Tensor],
) -> dict[str, torch.Tensor]:
    """
    Convert precessing spin parameters to lalsimulation parameters
    """
    mass_1, mass_2 = chirp_mass_and_mass_ratio_to_components(
        parameters["chirp_mass"], parameters["mass_ratio"]
    )

    parameters["mass_1"] = mass_1
    parameters["mass_2"] = mass_2

    # TODO: hard coding f_ref = 40 here b/c not sure best way to link this
    # to the f_ref specified in the config file
    incl, s1x, s1y, s1z, s2x, s2y, s2z = bilby_spins_to_lalsim(
        parameters["inclination"],
        parameters["phi_jl"],
        parameters["tilt_1"],
        parameters["tilt_2"],
        parameters["phi_12"],
        parameters["a_1"],
        parameters["a_2"],
        parameters["mass_1"],
        parameters["mass_2"],
        40,
        torch.zeros(len(mass_1), device=mass_1.device),
    )

    parameters["s1x"] = s1x
    parameters["s1y"] = s1y
    parameters["s1z"] = s1z
    parameters["s2x"] = s2x
    parameters["s2y"] = s2y
    parameters["s2z"] = s2z
    parameters["inclination"] = incl
    return parameters


def aligned_to_lalsimulation_parameters(
    parameters: dict[str, torch.Tensor],
) -> dict[str, torch.Tensor]:
    """
    Convert aligned spin parameters to lalsimulation parameters
    """
    mass_1, mass_2 = chirp_mass_and_mass_ratio_to_components(
        parameters["chirp_mass"], parameters["mass_ratio"]
    )

    parameters["mass_1"] = mass_1
    parameters["mass_2"] = mass_2

    parameters["s1x"] = torch.zeros_like(mass_1)
    parameters["s1y"] = torch.zeros_like(mass_1)

    parameters["s2x"] = torch.zeros_like(mass_1)
    parameters["s2y"] = torch.zeros_like(mass_1)

    parameters["s1z"] = parameters["chi1"]
    parameters["s2z"] = parameters["chi2"]
    return parameters
